import numpy as np
from scipy.integrate import quad


def totalTransformationEnthalpy(H_of_transformations, lower_temperature,
                                upper_temperature):
    '''
    Computes the total transformation enthalpy between two temperatures

    Parameters
    ----------
    H_of_transformations : dictionary or None or False
        If 'None' or 'False' then the function returs zero, as it is understood
        that no phase transition happens (or is known) for this substance. If
        this argument is a dictionary then each key of the dictionary must be
        an integer or float representing the temperature were the transition
        occurs (temperature is expected to be in Kelvin) and the value of each
        key must be the enthalpy of transformation during heating (expected to
        be in J/mol), represented as an integer or float. See notes for an
        example.

    lower_temperature : integer or float
        Lower temperature range (expected to be in Kelvin) to be considered for
        the calculation. If 'lower_temperature' is higher than
        'upper_temperature' the function raises a ValueError.

    upper_temperature : integer or float
        Upper temperature range (expected to be in Kelvin) to be considered for
        the calculation. If 'lower_temperature' is higher than
        'upper_temperature' the function raises a ValueError.

    Returns
    -------
    total_transformation_enthalpy : float
        Sum of all the transformation enthalpy that happens between
        'lower_temperature' and 'upper_temperature'.

    Notes
    -----
    See below an example of the dictionary format for 'H_of_transformation' for
    Silica (SiO2, see page 1505 in Ref. [1]).

    H_of_transformations_SiO2 = {
        847: 728.0,  # from low quartz to high quartz (J/mol)
        1079: 1995.8,  # from high quartz to high cristobalite (J/mol)
    }

    References
    ----------
    [1] Barin, I. (1995). Thermochemical data of pure substances.

    '''
    total_transformation_enthalpy = 0

    if isinstance(H_of_transformations, dict):
        if lower_temperature > upper_temperature:
            raise \
                ValueError('lower_temperature is higher than upper_temperature')

        for temperature in H_of_transformations:
            if lower_temperature <= temperature <= upper_temperature:
                total_transformation_enthalpy += \
                    H_of_transformations[temperature]

    return total_transformation_enthalpy


def excessEnthalpyFun(Cp_crystal_fun,
                      Cp_liquid_fun,
                      Tm,
                      delta_Hm,
                      H_transformations=None):
    '''
    Generate a function to compute the excess enthalpy.

    The excess enthalpy is the enthalpy of the liquid phase minus the enthalpy
    of the crystalline phase.

    Parameters
    ----------
    Cp_crystal_fun : callable
        Function with one argument for the temperature (in Kelvin) that returns
        the heat capacity at constant pressure for the crystalline phase.

    Cp_liquid_fun : callable
        Function with one argument for the temperature (in Kelvin) that returns
        the heat capacity at constant pressure for the liquid phase.

    Tm : integer or float
        Melting temperature.

    delta_Hm : integer or float
        Enthalpy of melting.

    H_transformations : dictionary or None or False
        If there are any crystalline phase transitions in the material, this
        information must be given as a dictionary were the keys are the
        temperature where the transition happens (in Kelvin) and the values are
        the enthalpy of transformations (in J/mol) during heating. See the
        documentation for the function totalTransformationEnthalpy for more
        details. Default value is None.

    Returns
    -------
    HexcFun : callable
        Function with one argument for temperature (in Kelvin) that returns the
        excess enthalpy at the given temperature. The excess enthalpy is the
        difference between the liquid and the crystalline enthalpy at the given
        temperature The argument of the function can be a scalar or an
        array-like.

    '''
    def fun(x):
        return Cp_liquid_fun(x) - Cp_crystal_fun(x)

    @np.vectorize
    def HexcFun(T):

        sum_transformation_enthalpy = \
            totalTransformationEnthalpy(H_transformations, T, Tm)
        integral = quad(fun, Tm, T, limit=100)[0]
        excess_enthalpy = delta_Hm + integral + sum_transformation_enthalpy

        return excess_enthalpy

    return HexcFun



import numpy as np
import warnings
from numpy import exp, log10
from copy import deepcopy as copy
from scipy.constants import N_A, pi, k
from scipy.integrate import odeint, solve_ivp
from scipy.misc import derivative
from scipy.optimize import brentq


def clusterPopGenOnlyMonomers(lenght, number_of_monomers=N_A, **kwargs):
    '''Generate a cluster population consisting of only monomers.

    Parameters
    ----------
    lenght : scalar
        Maximum number of particles possible in a cluster.

    number_of_monomers : scalar, optional
        Total number of monomers available. Default value is Avagadro's number.

    Returns
    -------
    N : array
        Array of size 'lenght' with all values being zero, except the first,
        for which is 'numebr_of_monomers'.

    '''
    N = np.zeros(lenght)
    N[0] = number_of_monomers
    return N


class KineticModelIsotropicSphere:
    '''Kinetic Model for isotropic sphere nucleation.

    Parameters
    ----------
    max_cluster_size : integer, optional
        Maximum number of particles possible in a cluster. Default value is
        500.

    min_cluster_size : integer, optional
        Minimum number of particles possible in a cluster. Default value is
        1.

    init_cluster_pop_function : callable, optional
        Function used to generate the initial cluster population distribution.
        The first argument of this function must be the maximum cluster size.
        All other arguments of the function are passed via
        'init_cluster_dictionary'. Default value is "clusterPopGenOnlyMonomers".

    init_cluster_dictionary : dictionary, optional
        Dictionaly with the arguments to pass to 'init_cluster_pop_function'.
        Default value is an empty dictionary.

    accuracy : 'cluster density' or 'cluster distribution' or None, optional
        It is not possible to be accurate in both the cluster density
        calculation and the cluster distribution calculation. If 'accuracy' is
        'cluster density' then the density calculation will be accurate. If
        'accuracy' is 'cluster distribution', then the cluster distribution
        will be accurate. If 'accuracy' is None then both Nv and cluster size
        distribution will be innacurate for cluster sizes close to the
        max_cluster_size. Default value is 'cluster density'

    References
    ----------
    [1] Kelton, K., and Greer, A.L. (2010). Nucleation in condensed matter:
        applications in materials and biology (Amsterdam: Elsevier).

    '''
    def __init__(self,
                 max_cluster_size=500,
                 min_cluster_size=1,
                 init_cluster_dictionary={},
                 init_cluster_pop_function=clusterPopGenOnlyMonomers,
                 accuracy='cluster density'):

        self.max_cluster_size = max_cluster_size
        self.min_cluster_size = min_cluster_size
        self.accuracy = accuracy
        initial_cluster_distribution = init_cluster_pop_function(
            self.max_cluster_size, **init_cluster_dictionary)
        self.cluster_distribution = [initial_cluster_distribution]
        self.time = [0]
        self.temperature = [np.nan]
        self.cluster_size = np.arange(self.min_cluster_size,
                                      self.max_cluster_size + 1)
        self.critical_cluster_size = [np.nan]
        self.supercritical_cluster_density = [np.nan]

    def W(self, cluster_size, driving_force, surface_energy, monomer_volume):
        '''Compute the work of formation of a cluster of size n.

        Parameters
        ----------
        cluster_size : scalar or array-like
            Number of monomers in a cluster.

        driving_force : float
            Thermodynamic driving force for nucleation. This value should be
            positive for temperatures below the melting temperature. Units of
            Joules per monomer.

        surface_energy : float
            Surface energy between a cluster and the ambient phase. Units of
            Joule per squared meter.

        monomer_volume : float
            Volume of a monomer. Units of cubic meter per monomer.

        Returns
        -------
        work_of_formation : scalar or array-like
            Work of formation of a cluster of size n. If n is an array, then
            this is the array of work of formation of all the clustes sizes in
            the array.

        Notes
        -----
        Equation for a formation of a spherical and isotropic cluster. See
        Equation (13) in  page 26 of Ref. [1].

        References
        ----------
        [1] Kelton, K., and Greer, A.L. (2010). Nucleation in condensed matter:
            applications in materials and biology (Amsterdam: Elsevier).

        '''
        surface = (36 * pi * (cluster_size * monomer_volume)**2)**(1 / 3)
        work_of_formation = surface_energy * surface \
            - cluster_size * driving_force
        return work_of_formation

    def _generate_matrix_K(self, temperature, diffusion_coeff, driving_force,
                           surface_energy, monomer_volume, jump_distance):
        '''Generate the K matrix.

        Parameters
        ----------
        temperature : scalar
            Absolute temperature.

        diffusion_coeff : scalar
            Diffusion coefficient.

        driving_force : float
            Thermodynamic driving force for nucleation. This value should be
            positive for temperatures below the melting temperature. Units of
            Joules per monomer.

        surface_energy : float
            Surface energy between a cluster and the ambient phase. Units of
            Joule per squared meter.

        monomer_volume : float
            Volume of a monomer. Units of cubic meter per monomer.

        jump_distance : float
            Average distance for an effective attachment or deattachment event.
            Units of meter.

        Returns
        -------
        K : array-like
            Matrix K used for the kinetic model.

        References
        ----------
        [1] Kelton, K., and Greer, A.L. (2010). Nucleation in condensed matter:
            applications in materials and biology (Amsterdam: Elsevier).

        '''
        n = np.arange(self.min_cluster_size - 1, self.max_cluster_size + 2)
        W = self.W(n, driving_force, surface_energy, monomer_volume)
        delW = W[1:] - W[:-1]
        O = 4 * n**(2 / 3)  # see pages 32 and 33 of Ref. [1]
        gamma = 6 * diffusion_coeff / jump_distance**2  # Eq. 30, page 33 of Ref. [1]

        kplus = O[1:-1] * gamma * exp(-delW[1:] / (2 * k * temperature))
        kminus = O[:-2] * gamma * exp(delW[:-1] / (2 * k * temperature))

        K = np.zeros((self.max_cluster_size - self.min_cluster_size + 1,
                      self.max_cluster_size - self.min_cluster_size + 1))
        i, j = np.indices(K.shape)

        K[i == j] = -(kplus + kminus)
        K[i == j - 1] = kminus[1:]
        K[i == j + 1] = kplus[:-1]

        if self.accuracy.lower() in ['cluster density', 'density']:
            K[-1, -1] += kplus[-1]
        elif self.accuracy.lower() in ['cluster distribution', 'distribution']:
            K[-1, -1] += kminus[-1]

        return K

    def computeCriticalClusterSize(self, driving_force, surface_energy,
                                   monomer_volume):
        '''Computes the cluster size (number of monomers) of a critical nucleus.

        A warning is raised if the critical cluster size is above the maximum
        cluster size.

        Parameters
        ----------
        driving_force : float
            Thermodynamic driving force for nucleation. This value should be
            positive for temperatures below the melting temperature. Units of
            Joules per monomer.

        surface_energy : float
            Surface energy between a cluster and the ambient phase. Units of
            Joule per squared meter.

        monomer_volume : float
            Volume of a monomer. Units of cubic meter per monomer.

        Returns
        -------
        critical_cluster_size : integer or NaN
            Number of monomers in a critical cluster size. Returns NaN if the
            critical cluster size is above the maximum cluster size.

        '''
        def Wfun(n):
            return self.W(n, driving_force, surface_energy, monomer_volume)

        def dW_dn(n):
            return derivative(Wfun, n, dx=1e-6)

        try:
            critical_cluster_size = brentq(dW_dn, self.min_cluster_size,
                                           self.max_cluster_size)
            critical_cluster_size = int(np.ceil(critical_cluster_size))

        except ValueError:
            msg = 'Critical cluster size is above the maximum cluster size'
            critical_cluster_size = np.nan
            warnings.warn(msg)

        return critical_cluster_size

    def _computeSupercriticalClusterDensity(self, cluster_distribution,
                                            critical_cluster_size,
                                            monomer_volume):
        '''Compute the supercritical cluster density.

        cluster_distribution : array-like
            This argument can be an array or array of arrays of the cluster
            distribution.

        critical_cluster_size : integer
            The size of a critical cluster.

        monomer_volume : float
            Volume of a monomer. Units of cubic meter per monomer.

        Returns
        -------
        supercritical_cluster_density : scalar or array-like
            Number of supercrcitical clusters per unity of volume.

        '''
        number_of_atoms = cluster_distribution.sum(axis=1)
        total_volume = monomer_volume * number_of_atoms
        critical_cluster_index = self.cluster_size.tolist().index(
            critical_cluster_size)
        supercritical_cluster_distribution = \
            cluster_distribution[:, critical_cluster_index:]
        supercritical_cluster_density = np.sum(
            np.round(supercritical_cluster_distribution),
            axis=1) / total_volume

        return supercritical_cluster_density

    def isothermalTreatment(self, temperature, time, diffusion_coeff,
                            surface_energy, driving_force, monomer_volume,
                            jump_distance):
        ''' Perform an isothermal treatment.

        Parameters
        ----------
        temperature : scalar
            Absolute temperature.

        time : float
            Total time of the isothermal treatment in seconds.

        diffusion_coeff : float
            Diffusion coefficient.

        driving_force : float
            Thermodynamic driving force for nucleation. This value should be
            positive for temperatures below the melting temperature. Units of
            Joules per monomer.

        surface_energy : float
            Surface energy between a cluster and the ambient phase. Units of
            Joule per squared meter.

        monomer_volume : float
            Volume of a monomer. Units of cubic meter per monomer.

        jump_distance : float
            Average distance for an effective attachment or deattachment event.
            Units of meter.

        References
        ----------
        [1] Kelton, K., and Greer, A.L. (2010). Nucleation in condensed matter:
            applications in materials and biology (Amsterdam: Elsevier).

        '''
        K = self._generate_matrix_K(temperature, diffusion_coeff,
                                    driving_force, surface_energy,
                                    monomer_volume, jump_distance)
        current_cluster_distribution = copy(self.cluster_distribution[-1])

        def dN_dt(N, t, K):
            return K @ N  # dot product of K and N

        time_resolution = max(10, int(np.ceil(log10(time) * 5)))
        times = np.logspace(0, log10(time), time_resolution)
        cluster_distribution_array = odeint(dN_dt,
                                            current_cluster_distribution,
                                            times,
                                            args=(K, ))

        critical_cluster_size = self.computeCriticalClusterSize(
            driving_force, surface_energy, monomer_volume)

        if np.isfinite(critical_cluster_size):
            supercritical_cluster_density = \
                self._computeSupercriticalClusterDensity(
                    cluster_distribution_array,
                    critical_cluster_size,
                    monomer_volume)
            self.critical_cluster_size.extend([critical_cluster_size] *
                                              time_resolution)
            self.supercritical_cluster_density.extend(
                supercritical_cluster_density)
        else:
            self.critical_cluster_size.extend([np.nan] * time_resolution)
            self.supercritical_cluster_density.extend([np.nan] *
                                                      time_resolution)

        self.time.extend(times + self.time[-1])
        self.temperature.extend([temperature] * time_resolution)
        self.cluster_distribution = np.concatenate(
            (self.cluster_distribution, cluster_distribution_array), axis=0)

    def __ramp(self, T1, T2, time, Dfun, sigmaFun, DGfun, lambdaFun,
               monomerVolFun, rampFun):
        pass



'''Classes for regression of viscosity data.'''

from scipy.stats import linregress
from lmfit import Model
from .equilibrium import logMYEGA


class _BaseViscosityRegression:
    '''
    Base class for viscosity regression.

    Parameters
    ----------

    table : pandas DataFrame, optional, must be a named argument
        DataFrame with a 'temperature' column and a 'log_viscosity' column. A
        RuntimeError is raised if the class is initiated without a table *or*
        without both temperature and log_viscosity arguments.

    temperature : array_like, optional, must be a named argument
        Temperature in Kelvin. A RuntimeError is raised if the class is initiated
        without a table *or* without both time and density arguments. If
        'table' is given then this argument is ignored.

    log_viscosity : array_like, optional, must be a named argument
        Base-10 logarithm of viscosity. It is highly recommended to use
        viscosity in units of Pascal second. A RuntimeError is raised if the
        class is initiated without a table *or* without both time and density
        arguments. If 'table' is given then this argument is ignored.

    '''
    def __init__(self, **kwargs):

        if 'table' in kwargs:

            table = kwargs['table']
            columns = table.columns.values

            if 'temperature' not in columns:
                name = self.__name__
                msg = f'The {name} class was initiated with a table with no temperature data'
                raise RuntimeError(msg)

            if 'log_viscosity' not in columns:
                name = self.__name__
                msg = f'The {name} class was initiated with a table with no log_viscosity data'
                raise RuntimeError(msg)

        elif 'temperature' in kwargs and 'log_viscosity' in kwargs:

            table = pd.DataFrame({
                'temperature': kwargs['temperature'],
                'log_viscosity': kwargs['log_viscosity'],
            })

        else:
            name = self.__name__
            msg = f'The {name} class was initiated with insuficient arguments'
            raise RuntimeError(msg)

        self.table = table

    def guess(self):
        '''
        Guess T12, fragility, and the logarithm of infinite viscosity.

        Returns
        -------
        guess_T12 : float
            Guess for the temperature were the log_10 of viscosity is 12

        guess_fragility : float
            Guess for the fragility index

        guess_log_eta_inf : float

        '''
        temperature = self.table['temperature']
        log_viscosity = self.table['log_viscosity']
        slope, intercept, _, _, _ = linregress(x=1 / temperature,
                                               y=log_viscosity)

        guess_T12 = slope / (12 - intercept)
        guess_T12 = guess_T12 if guess_T12 > 0 else min(temperature)
        guess_log_eta_inf = -3
        guess_fragility = 50

        model = Model(logMYEGA)
        fitresult = model.fit(log_viscosity,
                              T=temperature,
                              log_eta_inf=guess_log_eta_inf,
                              T12=guess_T12,
                              m=guess_fragility)

        guess_T12 = fitresult.params['T12'].value
        guess_fragility = fitresult.params['m'].value
        guess_log_eta_inf = fitresult.params['log_eta_inf'].value

        return guess_T12, guess_fragility, guess_log_eta_inf

    def _fit(self, model, weights=None, params=None, fitmethod='leastsq'):
        '''
        Regression of nucleation density data.

        Parameters
        ----------
        model : instance of lmfit's Model class
            Model to fit the data. The independent variable must be named 'T'.

        weights : array_like or None, optional
            The weights of log_viscosity to use during the regression. If None
            then no weights are applied. Default value is None.

        params : instance of lmfit's Parameters class or None, optional
            Optional Parameters instance to pass to the fit function. If None
            then the model will generate the Parameters class during fitting.
            Default value is None.

        fitmethod : str, optional
            Method to use for the regression. See lmfit's documentation for
            more information. Default value is 'leastsq'.

        Returns
        -------
        fitresult : instance of lmfit's ModelResult class
            Result of the regression. See lmfit for documentation on the
            ModelResult class.
        '''
        temperature = self.table['temperature']
        log_viscosity = self.table['log_viscosity']
        fitresult = model.fit(log_viscosity,
                              T=temperature,
                              method=fitmethod,
                              params=params,
                              weights=weights,
                              nan_policy='propagate')

        return fitresult


class MYEGA(_BaseViscosityRegression):
    '''
    Class for performing the MYEGA regression.

    Parameters
    ----------

    autofit : boolean, optional
        'True' if the regression should be performed during the Class initiation.
        'False' otherwise. Default value is True.

    table : pandas DataFrame, optional, must be a named argument
        DataFrame with a 'temperature' column and a 'log_viscosity' column. A
        RuntimeError is raised if the class is initiated without a table *or*
        without both temperature and log_viscosity arguments.

    temperature : array_like, optional, must be a named argument
        Temperature in Kelvin. A RuntimeError is raised if the class is initiated
        without a table *or* without both time and density arguments. If
        'table' is given then this argument is ignored.

    log_viscosity : array_like, optional, must be a named argument
        Base-10 logarithm of viscosity. It is highly recommended to use
        viscosity in units of Pascal second. A RuntimeError is raised if the
        class is initiated without a table *or* without both time and density
        arguments. If 'table' is given then this argument is ignored.

    '''
    def __init__(self, autofit=True, **kwargs):
        _BaseViscosityRegression.__init__(self, **kwargs)
        if autofit:
            self.fitresult, self.model = self.fit()

    def __str__(self):
        return 'MYEGA'

    def getModel(self, guess_T12, guess_fragility, guess_log_eta_inf):
        '''
        Creates a model for regression of the MYEGA equation.

        Parameters
        ----------
        guess_T12 : float
            Guess for the temperature were the viscosity is 10^12 Pa.s.

        guess_fragility : float
            Guess for the fragility index.

        guess_log_eta_inf : array_like, optional
            Guess for the base-10 logarithm of the infinite viscosity.

        Notes
        -----
        The parameters 'K' and 'C' from Eq. (6) from Ref. [1] are also added in
        the model paremeters.

        Returns
        -------
        model : instance of lmfit's Model class
            MYEGA model to fit experimental data.

        References
        ----------
        [1] Mauro, J.C., Yue, Y., Ellison, A.J., Gupta, P.K., and Allan, D.C.
            (2009). Viscosity of glass-forming liquids. Proceedings of the National
            Academy of Sciences of the United States of America 106, 19780–19784.

        '''
        model = Model(logMYEGA, name="MYEGA")

        model.set_param_hint('T12', vary=True, min=0, value=guess_T12)
        model.set_param_hint('m', vary=True, min=0, value=guess_fragility)
        model.set_param_hint('log_eta_inf',
                             vary=True,
                             max=11.99,
                             value=guess_log_eta_inf)
        model.set_param_hint(
            'K',
            vary=False,
            expr=r'(12-log_eta_inf)*T12*exp(1-m/(12-log_eta_inf))')
        model.set_param_hint('C',
                             vary=False,
                             expr=r'T12*(m/(12-log_eta_inf)-1)')

        return model

    def fit(self, weights=None, params=None, fitmethod='leastsq'):
        '''
        Regression of the viscosity data.

        Parameters
        ----------
        weights : array_like or None, optional
            The weights of log_viscosity to use during the regression. If None
            then no weights are applied. Default value is None.

        params : instance of lmfit's Parameters class or None, optional
            Optional Parameters instance to pass to the fit function. If None
            then the model will generate the Parameters class during fitting.
            Default value is None.

        fitmethod : str, optional
            Method to use for the regression. See lmfit's documentation for
            more information. Default value is 'leastsq'.

        Returns
        -------
        fitresult : instance of lmfit's ModelResult class
            Result of the regression. See lmfit for documentation on the
            ModelResult class.

        model : instance of lmfit's Model class
            MYEGA model to fit experimental data.

        '''
        guess_T12, guess_fragility, guess_log_eta_inf = self.guess()
        model = self.getModel(guess_T12, guess_fragility, guess_log_eta_inf)
        fitresult = self._fit(model, weights, params, fitmethod)

        return fitresult, model



'''Equations for equilibrium viscosity.'''

from numpy import exp, log


def logMYEGA(T, log_eta_inf, T12, m):
    """
    Computes the base-10 log of viscosity using the MYEGA equation.

    Parameters
    ----------
    T : float or array_like
        Temperature. Unit: Kelvin.

    log_eta_inf : float
        Base-10 logarithm of the asymptotic viscosity at the limit of infinite
        temperature.

    T12 : float
        Temperature were the viscosity is 10**12 Pa.s. Unit: Kelvin.

    m : float
        Fragility index as defined by Angell, see refs. [2]. Unitless.

    Returns
    -------
    log10_viscosity : float or array_like
        Returns the base-10 logarithm of viscosity.

    References
    ----------
    [1] Mauro, J.C., Yue, Y., Ellison, A.J., Gupta, P.K., and Allan, D.C.
        (2009). Viscosity of glass-forming liquids. Proceedings of the National
        Academy of Sciences of the United States of America 106, 19780–19784.

    [2] Angell, C.A. (1985). Strong and fragile liquids. In Relaxation in
        Complex Systems, K.L. Ngai, and G.B. Wright, eds. (Springfield: Naval
        Research Laboratory), pp. 3–12.

    """
    log10_viscosity = log_eta_inf + \
        (12-log_eta_inf)*(T12/T)*exp((m/(12-log_eta_inf)-1)*(T12/T-1))

    return log10_viscosity


def MYEGA(T, log_eta_inf, T12, m):
    """
    Computes the viscosity using the equation developed by Mauro and co-authors

    Parameters
    ----------
    T : float or array_like
        Temperature. Unit: Kelvin.

    log_eta_inf : float
        Base-10 logarithm of the asymptotic viscosity at the limit of infinite
        temperature.

    T12 : float
        Temperature were the viscosity is 10**12 Pa.s. Unit: Kelvin.

    m : float
        Fragility index as defined by Angell, see refs. [2]. Unitless.

    Returns
    -------
    viscosity : float or array_like
        Returns the viscosity in the units of eta_inf. Note: it is *not* the
        logarithm of viscosity.

    References
    ----------
    [1] Mauro, J.C., Yue, Y., Ellison, A.J., Gupta, P.K., and Allan, D.C.
        (2009). Viscosity of glass-forming liquids. Proceedings of the National
        Academy of Sciences of the United States of America 106, 19780–19784.

    [2] Angell, C.A. (1985). Strong and fragile liquids. In Relaxation in
        Complex Systems, K.L. Ngai, and G.B. Wright, eds. (Springfield: Naval
        Research Laboratory), pp. 3–12.

    """
    log10_viscosity = logMYEGA(T, log_eta_inf, T12, m)
    viscosity = 10**log10_viscosity

    return viscosity


def MYEGA_alt(T, eta_inf, K, C):
    """
    Computes the viscosity using the equation developed by Mauro and co-authors

    Parameters
    ----------
    T : float or array_like
        Temperature. Unit: Kelvin.

    eta_inf : float
        Asymptotic viscosity at the limit of infinite temperature.

    K : float
        See the original reference for the meaning. Unit: Kelvin.

    C : float
        See the original reference for the meaning. Unit: Kelvin.

    Returns
    -------
    viscosity : float or array_like
        Returns the viscosity in the units of eta_inf. Note: it is *not* the
        logarithm of viscosity.

    Notes
    -----
    In the original reference the equation is in base-10 logarithm, see Eq. (6)
    in [1]. To maintain the same meaning for the parameter K a log(10) is
    included in the double exponential expression here.

    References
    ----------
    [1] Mauro, J.C., Yue, Y., Ellison, A.J., Gupta, P.K., and Allan, D.C.
        (2009). Viscosity of glass-forming liquids. Proceedings of the National
        Academy of Sciences of the United States of America 106, 19780–19784.
    """
    viscosity = eta_inf * exp(log(10) * K / T * exp(C / T))

    return viscosity


def VFT(T, eta_inf, A, T0):
    """
    Computes the viscosity using the empirical Vogel-Fulcher-Tammann eq.

    Parameters
    ----------
    T : float or array_like
        Temperature. Unit: Kelvin.

    eta_inf : float
        Asymptotic viscosity at the limit of infinite temperature.

    A : float
        Adjustable parameter inside the exponential. Unit: Kelvin.

    T0 : float
        Divergence temperature. Unit: Kelvin.

    Returns
    -------
    viscosity : float or array_like
        Returns the viscosity in the units of eta_inf. Note: it is *not* the
        logarithm of viscosity.

    References
    ----------
    [1] Vogel, H. (1921). Das Temperatureabhängigketsgesetz der Viskosität von
        Flüssigkeiten. Physikalische Zeitschrift 22, 645–646.

    [2] Fulcher, G.S. (1925). Analysis of recent measurements of the viscosity
        of glasses. Journal of the American Ceramic Society 8, 339–355.

    [3] Tammann, G., and Hesse, W. (1926). Die Abhängigkeit der Viscosität von
        der Temperatur bie unterkühlten Flüssigkeiten. Z. Anorg. Allg. Chem.
        156, 245–257.
    """
    viscosity = eta_inf * exp(log(10) * A / (T - T0))
    return viscosity


def logVFT(T, log_eta_inf, A, T0):
    """
    Computes the base-10 log of viscosity using the Vogel-Fulcher-Tammann eq.

    Parameters
    ----------
    T : float or array_like
        Temperature. Unit: Kelvin.

    log_eta_inf : float
        Base-10 logarithm of the asymptotic viscosity at the limit of infinite
        temperature.

    A : float
        Adjustable parameter inside the exponential. Unit: Kelvin.

    T0 : float
        Divergence temperature. Unit: Kelvin.

    Returns
    -------
    log10_viscosity : float or array_like
        Returns the base-10 logarithm of viscosity.

    References
    ----------
    [1] Vogel, H. (1921). Das Temperatureabhängigketsgesetz der Viskosität von
        Flüssigkeiten. Physikalische Zeitschrift 22, 645–646.

    [2] Fulcher, G.S. (1925). Analysis of recent measurements of the viscosity
        of glasses. Journal of the American Ceramic Society 8, 339–355.

    [3] Tammann, G., and Hesse, W. (1926). Die Abhängigkeit der Viscosität von
        der Temperatur bie unterkühlten Flüssigkeiten. Z. Anorg. Allg. Chem.
        156, 245–257.
    """
    log10_viscosity = log_eta_inf + A / (T - T0)
    return log10_viscosity


def AM(T, eta_inf, alpha, beta):
    """
    Computes the viscosity using the equation developed by Avramov & Milchev.

    Parameters
    ----------
    T : float or array_like
        Temperature. Unit: Kelvin.

    eta_inf : float
        Asymptotic viscosity at the limit of infinite temperature.

    alpha : float
        Adjustable parameter, see original reference. Unitless.

    beta : float
        Adjustable parameter with unit of Kelvin.

    Returns
    -------
    viscosity : float or array_like
        Returns the viscosity in the units of eta_inf. Note: it is *not* the
        logarithm of viscosity.

    References
    ----------
    [1] Avramov, I., and Milchev, A. (1988). Effect of disorder on diffusion
        and viscosity in condensed systems. Journal of Non-Crystalline Solids
        104, 253–260.

    [2] Cornelissen, J., and Waterman, H.I. (1955). The viscosity temperature
        relationship of liquids. Chemical Engineering Science 4, 238–246.
    """
    viscosity = eta_inf * exp((beta / T)**alpha)
    return viscosity


def AG(T, eta_inf, B, S_conf_fun):
    """
    Computes the viscosity using the equation developed by Adam & Gibbs

    Parameters
    ----------
    T : float or array_like
        Temperature. Unit: Kelvin.

    eta_inf : float
        Asymptotic viscosity at the limit of infinite temperature.

    B : float
        Adjustable parameter related to the potential energy hindering the
        cooperative rearrangement per monomer segment.

    S_conf_fun : callable
        Function that computes the configurational entropy. This function
        accepts one argument, that is temperature.

    Returns
    -------
    viscosity : float or array_like
        Returns the viscosity in the units of eta_inf. Note: it is *not* the
        logarithm of viscosity.

    References
    ----------
    [1] Adam, G., and Gibbs, J.H. (1965). On the temperature dependence of
        cooperative relaxation properties in glass-forming liquids. The Journal
        of Chemical Physics 43, 139–146.
    """
    viscosity = eta_inf * exp(-B / (T * S_conf_fun(T)))

    return viscosity



import numpy as np
from scipy.integrate import quad
from .enthalpy import excessEnthalpyFun
from .entropy import excessEntropyFun


def excessGibbsFun(Cp_crystal_fun,
                   Cp_liquid_fun,
                   Tm,
                   delta_Hm,
                   H_transformations=None):
    '''
    Generate a function to compute the excess Gibbs free energy.

    The excess Gibbs free energy is the free energy of the liquid phase minus
    the free energy of the crystalline phase. The excess free energy, in the
    way it is defined here, must be positive below the melting point and
    negative above it. It is usual in the literature to call the excess Gibbs
    free energy as the crystallization driving force (not to be confused as the
    change in free energy due to crystallization, for which is just the
    negative of the excess Gibbs free energy).

    Parameters
    ----------
    Cp_crystal_fun : callable
        Function with one argument for the temperature (in Kelvin) that returns
        the heat capacity at constant pressure for the crystalline phase.

    Cp_liquid_fun : callable
        Function with one argument for the temperature (in Kelvin) that returns
        the heat capacity at constant pressure for the liquid phase.

    Tm : integer or float
        Melting temperature.

    delta_Hm : integer or float
        Enthalpy of melting.

    H_transformations : dictionary or None or False
        If there are any crystalline phase transitions in the material, this
        information must be given as a dictionary were the keys are the
        temperature where the transition happens (in Kelvin) and the values are
        the enthalpy of transformations (in J/mol) during heating. See the
        documentation for the function totalTransformationEntropy for more
        details. Default value is None.

    Returns
    -------
    GexcFun : callable
        Function with one argument for temperature (in Kelvin) that returns the
        excess Gibbs free energy at the given temperature. The excess Gibbs
        free energy is the difference between the liquid and the crystalline
        free energies at the given temperature The argument of the function can
        be a scalar or an array-like.

    '''
    HexcFun = excessEnthalpyFun(Cp_crystal_fun, Cp_liquid_fun, Tm, delta_Hm,
                                H_transformations)
    SexcFun = excessEntropyFun(Cp_crystal_fun, Cp_liquid_fun, Tm, delta_Hm,
                               H_transformations)

    def GexcFun(T):
        return HexcFun(T) - T * SexcFun(T)

    return GexcFun



import numpy as np
from numpy import exp
from scipy.constants import pi
from scipy.special import exp1


def wakeshima(time, steady_state_rate, time_lag):
    """
    Computes the nuclei density using the Wakeshima equation.

    Parameters
    ----------
    time : float or array_like
        Elapsed time.

    steady_state_rate : float
        Steady-state nucleation rate.

    time_lag : float
        Time-lag of transient nucleation. See ref. [1].

    Returns
    -------
    nuclei_density : float or array_like
        Returns the nuclei density.

    References
    ----------
    [1] Wakeshima, H. (1954). Time Lag in the Self‐Nucleation. The Journal of
        Chemical Physics 22, 1614–1615.
    """
    time_ratio = time/time_lag
    nuclei_density = steady_state_rate*(time_lag*(exp(-time_ratio) - 1) + time)
    return nuclei_density


def kashchiev(time, steady_state_rate, time_lag, time_shift=0,
              summation_ub=1000):
    """
    Computes the nuclei density using the Kashchiev equation.

    Parameters
    ----------
    time : float or array_like
        Elapsed time.

    steady_state_rate : float
        Steady-state nucleation rate.

    time_lag : float
        Time-lag of transient nucleation. See ref. [1].

    time_shift : float, optional
        Time-shift to account for double-stage treatments. See Eq. (33.65) in
        ref. [2]. Default value is 0.

    summation_ub : int, optional
        Upper boundary of the infinite summation. It is advisable to choose an
        even integer. Default value is 1000.

    Returns
    -------
    nuclei_density : float or array_like
        Returns the nuclei density.

    Notes
    -----
    This is the expression with the time-shift factor. For the original
    expression deduced in Ref. [1], set time_shift to zero.

    References
    ----------
    [1] Kashchiev, D. (1969). Solution of the non-steady state problem in
        nucleation kinetics. Surface Science 14, 209–220.

    [2] Kashchiev, D. (2000). Nucleation basic theory with applications.
    """
    if time_lag == 0:
        nuclei_density = steady_state_rate*time

    else:

        @np.vectorize
        def _kashchiev(t):
            if t <= time_shift:
                return 0

            time_ratio = (t - time_shift)/time_lag

            def summationParticle(n):
                return ((-1)**(n%2))*exp(-n**2*time_ratio)/n**2

            summation = np.sum(summationParticle(np.arange(1, summation_ub + 1)))
            N = steady_state_rate*time_lag*(time_ratio - pi**2/6 - 2*summation)
            return N if N > 0 else 0

        nuclei_density = _kashchiev(time)

    return nuclei_density


def kashchievMasterCurve(time_ratio, summation_ub=1000):
    """
    Computes the normalized nuclei density using the Kashchiev equation.

    Parameters
    ----------
    time_ratio : float or array_like
        The ratio between the difference of time and time-shift over time-lag.

    summation_ub : int, optional
        Upper boundary of the infinite summation. It is advisable to choose an
        even integer. Default value is 1000.

    Returns
    -------
    normalized_nuclei_density : float or array_like
        Returns the normalized nuclei density, which is the nuclei density over
        the product of the steady-state rate and the time-lag.

    References
    ----------
    [1] Kashchiev, D. (1969). Solution of the non-steady state problem in
        nucleation kinetics. Surface Science 14, 209–220.

    [2] Kashchiev, D. (2000). Nucleation basic theory with applications.
    """
    @np.vectorize
    def _kashchiev(time_ratio):

        def summationParticle(n):
            return ((-1)**(n%2))*exp(-n**2*time_ratio)/n**2

        summation = np.sum(summationParticle(np.arange(1, summation_ub + 1)))
        Normalized_N = time_ratio - pi**2/6 - 2*summation
        return Normalized_N if Normalized_N > 0 else 0

    normalized_nuclei_density = _kashchiev(time_ratio)
    return normalized_nuclei_density


def shneidman(time, steady_state_rate, time_lag, incubation_time):
    """
    Computes the nuclei density using the Shneidman equation.

    Parameters
    ----------
    time : float or array_like
        Elapsed time.

    steady_state_rate : float
        Steady-state nucleation rate.

    time_lag : float
        Time-lag of transient nucleation. See Ref. [1].

    incubation_time : float
        Incubation time due to the double-stage treatment. See Ref. [1].

    Returns
    -------
    nuclei_density : float or array_like
        Returns the nuclei density.

    References
    ----------
    [1] Shneidman, V.A. (1988). Establishment of a steady-state nucleation
        regime. Theory and comparison with experimental data for glasses. Sov.
        Phys. Tech. Phys. 33, 1338–1342.
    """

    if time_lag == 0:
        nuclei_density = steady_state_rate*time

    else:

        time_ratio = (time - incubation_time)/time_lag
        nuclei_density = time_lag*steady_state_rate*exp1(exp(-time_ratio))

    return nuclei_density



'''Equations for computing the effective diffusion coefficient from
viscosity.'''

from scipy.constants import k, pi


def diffCoeffEyring(T, viscosity, diameter):
    """
    Computes the viscosity diffusion coefficient using Eyring equation

    Parameters
    ----------
    T : float or array_like
        Temperature. Unit: Kelvin.

    viscosity : float or array_like with same lenght as T
        Viscosity at temperature T.

    diameter : float or array_like with same lenght as T
        The diameter of the structural unit that is moving due to viscous flow.

    Returns
    -------
    out : float or array_like
        Returns the effective diffusion coefficient computed using the Eyring
        equation. This equation is similar to the Stokes-Einstein equation, but
        they were obtained by different routes.

    References
    ----------
    [1] Eyring, H. (1936). Viscosity, plasticity, and diffusion as examples of
        absolute reaction rates. The Journal of Chemical Physics 4, 283–291.
    """
    return k*T/(viscosity*diameter)


def diffCoeffStokesEinstein(T, viscosity, diameter):
    """
    Computes the viscosity diffusion coefficient using Stokes-Einstein equation

    Parameters
    ----------
    T : float or array_like
        Temperature. Unit: Kelvin.

    viscosity : float or array_like with same lenght as T
        Viscosity at temperature T.

    diameter : float or array_like with same lenght as T
        The diameter of the structural unit that is moving due to viscous flow.

    Returns
    -------
    out : float or array_like
        Returns the effective diffusion coefficient computed using the
        Stokes-Einstein equation. This equation is similar to the Eyring
        equation, but they were obtained by different routes.

    References
    ----------
    [1] Einstein, A. (1905). On the movement of small particles suspended in
        stationary liquids required by the molecular-kinetic theory of heat.
        Annalen Der Physik 17, 549–560.

    [2] Einstein, A. (1905). Über die von der molekularkinetischen Theorie der
        Wärme geforderte Bewegung von in ruhenden Flüssigkeiten suspendierten
        Teilchen. Annalen Der Physik 322, 549–560.

    [3] Stokes, G.G. (1851). On the effect of the internal friction of fluids
        on the motion of pendulums. Transactions of the Cambridge Philosophical
        Society 9, 8–106.
    """
    return k*T/(3*pi*viscosity*diameter)



'''Classes for regression of nucleation density data.'''

import pandas as pd
import numpy as np
from numpy import log10
from scipy.stats import theilslopes
from scipy.constants import pi
from lmfit import Model, Parameters

from .density import wakeshima, kashchiev, shneidman


class _BaseDensityRegression:
    '''
    Base class for density regression.

    Parameters
    ----------

    table : pandas DataFrame, optional, must be a named argument
        DataFrame with a 'time' and a 'density' column. If an 'ID' column is
        also present, then each unique ID is considered a separate dataset. A
        RuntimeError is raised if the class is initiated without a table *or*
        without both time and density arguments.

    time : array_like, optional, must be a named argument
        Elapsed time. A RuntimeError is raised if the class is initiated
        without a table *or* without both time and density arguments. If
        'table' is given then this argument is ignored.

    density : array_like, optional, must be a named argument
        Nuclei density. A RuntimeError is raised if the class is initiated
        without a table *or* without both time and density arguments. If
        'table' is given then this argument is ignored.

    ID : array_like, optional, must be a named argument
        The ID for each of the 'time' and 'density' data. Each unique ID is
        treated as a separate dataset. If 'table' is given then this argument
        is ignored.
    '''
    def __init__(self, **kwargs):

        if 'table' in kwargs:

            table = kwargs['table']
            columns = table.columns.values

            if 'time' not in columns:
                name = self.__name__
                msg = f'The {name} class was initiated with a table with no "time" column'
                raise RuntimeError(msg)

            if 'density' not in columns:
                name = self.__name__
                msg = f'The {name} class was initiated with a table with no "density" column'
                raise RuntimeError(msg)

            if 'ID' not in columns:
                table['ID'] = np.zeros(len(table)).astype(bool)

        elif 'time' in kwargs and 'density' in kwargs:

            table = pd.DataFrame({
                'time': kwargs['time'],
                'density': kwargs['density'],
            })

            if 'ID' in kwargs:
                table['ID'] = kwargs['ID']
            else:
                table['ID'] = np.zeros(len(table)).astype(bool)

            if 'temperature' in kwargs:
                table['temperature'] = kwargs['temperature']

        else:
            name = self.__name__
            msg = f'The {name} class was initiated with insuficient arguments'
            raise RuntimeError(msg)

        self.table = table

    def dataGenerator(self):
        '''
        Generator that yields each individual dataset

        Yields
        ------
        ID
            The identification of the yielded dataset

        time : array_like
            Elapsed time of the yielded dataset

        density : array_like
            Nuclei density of the yielded dataset

        table_slice : pandas DataFrame
            An instance of the table DataFrame of the yielded dataset
        '''
        for ID in self.table['ID'].unique():

            logic = self.table['ID'] == ID
            table_slice = self.table[logic]
            time = table_slice['time'].values
            density = table_slice['density'].values

            yield ID, time, density, table_slice

    def guess(self, time, density):
        '''
        Guess the steady-state nucleation rate and induction time

        The guess is based on a robust linear regression of the density over
        time data. Guesses are useful for non-linear regressions.

        Parameters
        ----------
        time : array_like
            Elapsed time.

        density : array_like
            Nuclei density.

        Returns
        -------
        guess_rate : float
            Guess for the steady-state nucleatio rate for the yielded dataset

        guess_induction_time : float
            Guess for the induction time for the yielded dataset
        '''
        slope, intercept, _, _ = theilslopes(x=time, y=density)
        guess_rate, guess_induction_time = slope, -intercept / slope

        if guess_induction_time < 1e-3:
            guess_induction_time = 1e-3

        return guess_rate, guess_induction_time

    def dataGeneratorWithGuess(self):
        '''
        Generate individual datasets and guesses for nucl. rate and induct. time

        Yields
        ------
        ID
            The identification of the yielded dataset

        time : array_like
            Elapsed time of the yielded dataset

        density : array_like
            Nuclei density of the yielded dataset

        guess_rate : float
            Guess for the steady-state nucleatio rate for the yielded dataset

        guess_induction_time : float
            Guess for the induction time for the yielded dataset
        '''
        for ID in self.table['ID'].unique():

            logic = self.table['ID'] == ID
            table_slice = self.table[logic]
            time = table_slice['time'].values
            density = table_slice['density'].values

            guess_rate, guess_induction_time = self.guess(time, density)

            yield ID, time, density, guess_rate, guess_induction_time

    def _fit(self,
             model,
             time,
             density,
             density_weights=None,
             params=None,
             fitmethod='leastsq'):
        '''
        Regression of nucleation density data.

        Parameters
        ----------
        model : instance of lmfit's Model class
            Model to fit the data. The independent variable must be named
            'time'.

        time : array_like
            Elapsed time.

        density : array_like
            Nuclei density.

        density_weights : array_like or None, optional
            The weights of density to use during the regression. Default value
            is None.

        params : instance of lmfit's Parameters class or None, optional
            Optional Parameters instance to pass to the fit function. If None
            then the model will generate the Parameters class during fitting.
            Default value is None.

        fitmethod : str, optional
            Method to use for the regression. See lmfit's documentation for
            more information. Default value is 'leastsq'.

        Returns
        -------
        fitresult : instance of lmfit's ModelResult class
            Result of the regression. See lmfit for documentation on the
            ModelResult class.
        '''
        fitresult = model.fit(density,
                              time=time,
                              method=fitmethod,
                              params=params,
                              weights=density_weights,
                              nan_policy='propagate')

        return fitresult

    def _fitBrute(self,
                  model,
                  time,
                  density,
                  time_weights=None,
                  params=None,
                  fitmethod='leastsq'):
        '''
        Brute-force regression of nucleation density data.

        Some combinations of dataset and models are quite difficult to fit.
        This method is a brute-force approach to find an answer that converges.

        Parameters
        ----------
        model : instance of lmfit's Model class
            Model to fit the data. The independent variable must be named
            'time'.

        time : array_like
            Elapsed time.

        density : array_like
            Nuclei density.

        density_weights : array_like or None, optional
            The weights of density to use during the regression. Default value
            is None.

        params : instance of lmfit's Parameters class or None, optional
            Optional Parameters instance to pass to the fit function. If None
            then the model will generate the Parameters class during fitting.
            Default value is None.

        fitmethod : str, optional
            Method to use for the final regression. See lmfit's documentation
            for more information. Default value is 'leastsq'.

        Returns
        -------
        fitresult : instance of lmfit's ModelResult class
            Result of the regression. See lmfit for documentation on the
            ModelResult class.
        '''
        fitresult = self._fit(model,
                              time,
                              density,
                              time_weights,
                              params,
                              fitmethod='brute')

        fitresult = self._fit(model,
                              time,
                              density,
                              time_weights,
                              fitresult.params,
                              fitmethod='differential_evolution')

        fitresult = self._fit(model,
                              time,
                              density,
                              time_weights,
                              fitresult.params,
                              fitmethod=fitmethod)

        return fitresult

    def __confidenceInterval(self, fitresult, ci_names, ci_sigmas=(0.95, )):
        # TODO
        pass


class Wakeshima(_BaseDensityRegression):
    '''
    Class for density regression considering the Wakeshima equation.

    Parameters
    ----------

    table : pandas DataFrame, optional, must be a named argument
        DataFrame with a 'time' and a 'density' column. If an 'ID' column is
        also present, then each unique ID is considered a separate dataset. A
        RuntimeError is raised if the class is initiated without a table *or*
        without both time and density arguments.

    time : array_like, optional, must be a named argument
        Elapsed time. A RuntimeError is raised if the class is initiated
        without a table *or* without both time and density arguments. If
        'table' is given then this argument is ignored.

    density : array_like, optional, must be a named argument
        Nuclei density. A RuntimeError is raised if the class is initiated
        without a table *or* without both time and density arguments. If
        'table' is given then this argument is ignored.

    ID : array_like, optional, must be a named argument
        The ID for each of the 'time' and 'density' data. Each unique ID is
        treated as a separate dataset. If 'table' is given then this argument
        is ignored.

    References
    ----------
    [1] Wakeshima, H. (1954). Time Lag in the Self‐Nucleation. The Journal of
        Chemical Physics 22, 1614–1615.
    '''
    def getModel(self, guess_rate, guess_induction_time):
        '''
        Creates a model for regression of the Wakeshima equation.

        Parameters
        ----------
        guess_rate : float
            Guess for the steady-state nucleation rate.

        guess_induction_time : float
            Guess for the nucleation induction time.

        Returns
        -------
        model : instance of lmfit's Model class
            Wakeshima model to fit experimental data.
        '''
        model = Model(wakeshima)

        model.set_param_hint('log_rate',
                             vary=True,
                             max=25,
                             value=log10(guess_rate))
        model.set_param_hint('steady_state_rate',
                             vary=False,
                             expr=r'10**log_rate')
        model.set_param_hint('log_induction_time',
                             vary=True,
                             value=log10(guess_induction_time))
        model.set_param_hint('time_lag',
                             vary=False,
                             expr=r'10**log_induction_time')
        model.set_param_hint('induction_time',
                             vary=False,
                             expr=r'10**log_induction_time')

        return model

    def fit(self,
            time,
            density,
            density_weights=None,
            params=None,
            fitmethod='leastsq'):
        '''
        Regression of nucleation density data.

        Parameters
        ----------
        time : array_like
            Elapsed time.

        density : array_like
            Nuclei density.

        density_weights : array_like or None, optional
            The weights of density to use during the regression. Default value
            is None.

        params : instance of lmfit's Parameters class or None, optional
            Optional Parameters instance to pass to the fit function. If None
            then the model will generate the Parameters class during fitting.
            Default value is None.

        fitmethod : str, optional
            Method to use for the regression. See lmfit's documentation for
            more information. Default value is 'leastsq'.

        Returns
        -------
        fitresult : instance of lmfit's ModelResult class
            Result of the regression. See lmfit for documentation on the
            ModelResult class.

        model : instance of lmfit's Model class
            Wakeshima model to fit experimental data.
        '''
        guess_rate, guess_induction_time = self.guess(time, density)
        model = self.getModel(guess_rate, guess_induction_time)
        fitresult = self._fit(model, time, density, density_weights, params,
                              fitmethod)

        return fitresult, model


class Kashchiev(_BaseDensityRegression):
    '''
    Class for density regression considering the Kashchiev equation.

    Parameters
    ----------

    table : pandas DataFrame, optional, must be a named argument
        DataFrame with a 'time' and a 'density' column. If an 'ID' column is
        also present, then each unique ID is considered a separate dataset. A
        RuntimeError is raised if the class is initiated without a table *or*
        without both time and density arguments.

    time : array_like, optional, must be a named argument
        Elapsed time. A RuntimeError is raised if the class is initiated
        without a table *or* without both time and density arguments. If
        'table' is given then this argument is ignored.

    density : array_like, optional, must be a named argument
        Nuclei density. A RuntimeError is raised if the class is initiated
        without a table *or* without both time and density arguments. If
        'table' is given then this argument is ignored.

    ID : array_like, optional, must be a named argument
        The ID for each of the 'time' and 'density' data. Each unique ID is
        treated as a separate dataset. If 'table' is given then this argument
        is ignored.

    summation_ub : int, optional
        Upper boundary of the infinite summation. It is advisable to choose an
        even integer. Default value is 1000.

    use_time_shift : bool, optional
        If False then the original Kashchiev equation is used (equivalent to
        setting the time_shift to zero). If True then the expression with a
        time-shift will be considered for the regression. The time-shift is a
        parameter related to the dissolution of supercritical nuclei in a
        double-stage nucleation treatment. Default value is False.

    References
    ----------
    [1] Kashchiev, D. (1969). Solution of the non-steady state problem in
        nucleation kinetics. Surface Science 14, 209–220.

    [2] Kashchiev, D. (2000). Nucleation basic theory with applications.
    '''
    def __init__(self, **kwargs):
        _BaseDensityRegression.__init__(self, **kwargs)
        self.summation_ub = kwargs.get('summation_ub', 1000)
        self.use_time_shift = kwargs.get('use_time_shift', False)

    def __str__(self):
        if self.use_time_shift:
            time_shift_info = ' considering the time-shift'
        else:
            time_shift_info = ''
        return f'Kaschiev{time_shift_info} with {self.summation_ub} summation terms'

    def getModel(self, guess_rate, guess_induction_time, time=None):
        '''
        Creates a model for regression of the Kashchiev equation.

        Parameters
        ----------
        guess_rate : float
            Guess for the steady-state nucleation rate.

        guess_induction_time : float
            Guess for the nucleation induction time.

        time : array_like, optional
            Elapsed time. Only necessary if self.use_time_shift is True. Used
            to give a maximum value for the time_shift.

        Returns
        -------
        model : instance of lmfit's Model class
            Kashchiev model to fit experimental data.
        '''
        model = Model(kashchiev)

        model.set_param_hint('log_rate',
                             vary=True,
                             min=5,
                             max=25,
                             value=log10(guess_rate),
                             brute_step=1)
        model.set_param_hint('steady_state_rate',
                             vary=False,
                             expr=r'10**log_rate')
        model.set_param_hint('log_time_lag',
                             vary=True,
                             min=-1,
                             max=9,
                             value=log10(guess_induction_time * 6 / pi**2),
                             brute_step=0.5)
        model.set_param_hint('time_lag', vary=False, expr=r'10**log_time_lag')
        model.set_param_hint('summation_ub',
                             vary=False,
                             value=self.summation_ub)

        if self.use_time_shift:
            time_shift_guess = min(min(time), guess_induction_time) / 2

            model.set_param_hint('log_time_shift',
                                 vary=True,
                                 min=-1,
                                 max=log10(max(time)),
                                 value=log10(time_shift_guess),
                                 brute_step=0.5)
            model.set_param_hint('time_shift',
                                 vary=False,
                                 expr=r'10**log_time_shift')
        else:
            model.set_param_hint('time_shift', vary=False, value=0)

        model.set_param_hint('induction_time',
                             vary=False,
                             expr='time_lag*pi**2/6 + time_shift')
        return model

    def fit(self,
            time,
            density,
            density_weights=None,
            params=None,
            fitmethod='leastsq',
            time_shift_threshold=0,
            time_lag_threshold=0):
        '''
        Regression of nucleation density data.

        Parameters
        ----------
        time : array_like
            Elapsed time.

        density : array_like
            Nuclei density.

        density_weights : array_like or None, optional
            The weights of density to use during the regression. Default value
            is None.

        params : instance of lmfit's Parameters class or None, optional
            Optional Parameters instance to pass to the fit function. If None
            then the model will generate the Parameters class during fitting.
            Default value is None.

        fitmethod : str, optional
            Method to use for the regression. See lmfit's documentation for
            more information. Default value is 'leastsq'.

        time_shift_threshold : float, optional
            If the time_shift obtained from the regression is lower than the
            threshold value, then the time_shift is fixed at zero and a new
            regression is done. Default value is zero.

        time_shift_threshold : float, optional
            If the time_lag obtained from the regression is lower than the
            threshold value, then the time_lag is fixed at zero and a new
            regression is done. Default value is zero.

        Notes
        -----
        When considering the Kashchiev equation with time-shift, the fitting
        procedure checks if the standard deviation of the time-shift is greater
        than the actual value of the time-shift. If so, then the time-shift is
        forced to zero.

        Returns
        -------
        fitresult : instance of lmfit's ModelResult class
            Result of the regression. See lmfit for documentation on the
            ModelResult class.

        model : instance of lmfit's Model class
            Kashchiev model to fit experimental data.
        '''
        guess_rate, guess_induction_time = self.guess(time, density)
        model = self.getModel(guess_rate, guess_induction_time, time)
        fitresult = self._fit(model, time, density, density_weights, params,
                              fitmethod)

        if self.use_time_shift:
            fitparams = fitresult.params
            time_shift = fitparams['time_shift'].value
            time_shift_std = fitparams['time_shift'].stderr
            if time_shift < time_shift_threshold or time_shift_std > time_shift:
                fitparams.add('time_shift', vary=False, value=0)
                fitparams.add('log_time_shift', vary=False, value=-np.inf)
                fitresult = self._fit(model, time, density, density_weights,
                                      fitparams, fitmethod)

        fitparams = fitresult.params
        time_lag = fitparams['time_lag'].value
        time_lag_std = fitparams['time_lag'].stderr
        if time_lag < time_lag_threshold:
            fitparams.add('time_lag', vary=False, value=0)
            fitparams.add('log_time_lag', vary=False, value=-np.inf)
            fitresult = self._fit(model, time, density, density_weights,
                                  fitparams, fitmethod)

        return fitresult, model


class __Shneidman(_BaseDensityRegression):
    '''
    Class for density regression considering the Shneidman equation.

    Parameters
    ----------

    table : pandas DataFrame, optional, must be a named argument
        DataFrame with a 'time' and a 'density' column. If an 'ID' column is
        also present, then each unique ID is considered a separate dataset. A
        RuntimeError is raised if the class is initiated without a table *or*
        without both time and density arguments.

    time : array_like, optional, must be a named argument
        Elapsed time. A RuntimeError is raised if the class is initiated
        without a table *or* without both time and density arguments. If
        'table' is given then this argument is ignored.

    density : array_like, optional, must be a named argument
        Nuclei density. A RuntimeError is raised if the class is initiated
        without a table *or* without both time and density arguments. If
        'table' is given then this argument is ignored.

    ID : array_like, optional, must be a named argument
        The ID for each of the 'time' and 'density' data. Each unique ID is
        treated as a separate dataset. If 'table' is given then this argument
        is ignored.

    detectable_radius : float, optional
        The size of the detectable radius. See ref. [1] for more details. The
        default value is 1e-6.

    time_ratio_fun : function or None, optional
        The function to compute the ratio between the incubation time and the
        time-lag. The function must accept two key arguments: steady_state_rate
        and time_lag. If, instead, time_ratio_fun is None then the adjustable
        parameters will be considered free and independent. Default value is
        None.

    References
    ----------
    [1] Shneidman, V.A. (1988). Establishment of a steady-state nucleation
        regime. Theory and comparison with experimental data for glasses. Sov.
        Phys. Tech. Phys. 33, 1338–1342.
    '''
    def __init__(self, **kwargs):
        _BaseDensityRegression.__init__(self, **kwargs)
        self.detectable_radius = kwargs.get('detectable_radius', 1e-6)
        self.time_ratio_fun = kwargs.get('time_ratio_fun', None)

    def __str__(self):
        if self.time_ratio_fun:
            return 'Shneidman with 2 free adjustable parameteres'
        else:
            return 'Shneidman with 3 free adjustable parameters'

    def getModel(self, guess_rate, guess_induction_time):
        '''
        Creates a model for regression of the Shneidman equation.

        Parameters
        ----------
        guess_rate : float
            Guess for the steady-state nucleation rate.

        guess_induction_time : float
            Guess for the nucleation induction time.

        time : array_like, optional
            Elapsed time. Only necessary if self.use_time_shift is True. Used
            to give a maximum value for the time_shift.

        Returns
        -------
        model : instance of lmfit's Model class
            Kashchiev model to fit experimental data.
        '''
        model = Model(shneidman)

        params = Parameters(
            usersyms={
                'euler_gamma': np.euler_gamma,
                'time_ratio_fun': self.time_ratio_fun,
                'detectable_radius': self.detectable_radius,
            })

        params.add('log_rate',
                   vary=True,
                   min=5,
                   max=25,
                   value=log10(guess_rate),
                   brute_step=1)
        params.add('steady_state_rate', vary=False, expr=r'10**log_rate')
        params.add('log_time_lag',
                   vary=True,
                   min=-1,
                   max=9,
                   value=log10(guess_induction_time / (2 * np.euler_gamma)),
                   brute_step=0.5)
        params.add('time_lag', vary=False, expr=r'10**log_time_lag')

        if self.time_ratio_fun:
            expr = r'time_lag*time_ratio_fun(steady_state_rate, time_lag, detectable_radius)'
            params.add('incubation_time', vary=False, expr=expr)

        else:
            params.add('log_incubation_time',
                       vary=True,
                       min=-1,
                       max=9,
                       value=log10(guess_induction_time / 2),
                       brute_step=0.5)
            params.add('incubation_time',
                       vary=False,
                       expr=r'10**log_incubation_time')

        params.add('induction_time',
                   vary=False,
                   expr=r'incubation_time + euler_gamma*time_lag')

        return model, params

    def fit(self,
            time,
            density,
            density_weights=None,
            params=None,
            fitmethod='leastsq',
            time_shift_threshold=0):
        '''
        Regression of nucleation density data.

        Parameters
        ----------
        time : array_like
            Elapsed time.

        density : array_like
            Nuclei density.

        density_weights : array_like or None, optional
            The weights of density to use during the regression. Default value
            is None.

        params : instance of lmfit's Parameters class or None, optional
            Optional Parameters instance to pass to the fit function. If None
            then the model will generate the Parameters class during fitting.
            Default value is None.

        fitmethod : str, optional
            Method to use for the regression. See lmfit's documentation for
            more information. Default value is 'leastsq'.

        time_shift_threshold : float, optional
            If the time_shift obtained from the regression is lower than the
            threshold value, then the time_shift is fixed at zero and a new
            regression is done. Default value is zero.

        Returns
        -------
        fitresult : instance of lmfit's ModelResult class
            Result of the regression. See lmfit for documentation on the
            ModelResult class.

        model : instance of lmfit's Model class
            Shneidman model to fit experimental data.
        '''
        guess_rate, guess_induction_time = self.guess(time, density)
        model, params_ = self.getModel(guess_rate, guess_induction_time)

        if params is None:
            params = params_

        fitresult = self._fit(model, time, density, density_weights, params,
                              fitmethod)

        fitparams = fitresult.params
        incubation_time = fitparams['incubation_time'].value
        if incubation_time < 1:
            fitparams.add('incubation_time', vary=False, value=0)
            fitparams.add('log_incubation_time', vary=False, value=-np.inf)
            fitresult = self._fit(model, time, density, density_weights,
                                  fitparams, fitmethod)

        fitparams = fitresult.params
        time_lag = fitparams['time_lag'].value
        if time_lag < 1:
            fitparams.add('time_lag', vary=False, value=0)
            fitparams.add('log_time_lag', vary=False, value=-np.inf)
            fitresult = self._fit(model, time, density, density_weights,
                                  fitparams, fitmethod)

        return fitresult, model



import numpy as np
from scipy.integrate import quad
from scipy.optimize import brentq


def totalTransformationEntropy(H_of_transformations, lower_temperature,
                               upper_temperature):
    '''
    Computes the total transformation entropy between two temperatures

    Parameters
    ----------
    H_of_transformations : dictionary or None or False
        If 'None' or 'False' then the function returs zero, as it is understood
        that no phase transition happens (or is known) for this substance. If
        this argument is a dictionary then each key of the dictionary must be
        an integer or float representing the temperature were the transition
        occurs (temperature is expected to be in Kelvin) and the value of each
        key must be the enthalpy of transformation during heating (expected to
        be in J/mol), represented as an integer or float. See notes for an
        example.

    lower_temperature : integer or float
        Lower temperature range (expected to be in Kelvin) to be considered for
        the calculation. If 'lower_temperature' is higher than
        'upper_temperature' the function raises a ValueError.

    upper_temperature : integer or float
        Upper temperature range (expected to be in Kelvin) to be considered for
        the calculation. If 'lower_temperature' is higher than
        'upper_temperature' the function raises a ValueError.

    Returns
    -------
    total_transformation_entropy : float
        Sum of all the transformation entropy that happens between
        'lower_temperature' and 'upper_temperature'.

    Notes
    -----
    See below an example of the dictionary format for 'H_of_transformation' for
    Silica (SiO2, see page 1505 in Ref. [1]).

    H_of_transformations_SiO2 = {
        847: 728.0,  # from low quartz to high quartz (J/mol)
        1079: 1995.8,  # from high quartz to high cristobalite (J/mol)
    }

    References
    ----------
    [1] Barin, I. (1995). Thermochemical data of pure substances.

    '''
    total_transformation_entropy = 0

    if isinstance(H_of_transformations, dict):
        if lower_temperature > upper_temperature:
            raise \
                ValueError('lower_temperature is higher than upper_temperature')

        for temperature in H_of_transformations:
            if lower_temperature <= temperature <= upper_temperature:
                total_transformation_entropy += \
                    H_of_transformations[temperature] / temperature

    return total_transformation_entropy


def entropyCrystalFun(Cp_crystal_fun, H_transformations=None):
    '''
    Generate a function to compute the entropy of a crystal.

    Parameters
    ----------
    Cp_crystal_fun : callable
        Function with one argument for the temperature (in Kelvin) that returns
        the heat capacity at constant pressure for the crystalline phase. It is
        necessary that Cp_crystal_fun returns finite heat capacity values
        between zero Kelvin and any temperature that will be considerade to
        compute the entropy of the crystal.

    H_transformations : dictionary or None or False
        If there are any crystalline phase transitions in the material, this
        information must be given as a dictionary were the keys are the
        temperature where the transition happens (in Kelvin) and the values are
        the enthalpy of transformations (in J/mol) during heating. See the
        documentation for the function totalTransformationEntropy for more
        details. Default value is None.

    Returns
    -------
    ScrystalFun : callable
        Function with one argument for temperature (in Kelvin) that returns the
        entropy of the crystalline phase at the given temperature. The argument
        of the function can be a scalar or an array-like.

    '''
    def fun(x):
        return Cp_crystal_fun(x) / x

    @np.vectorize
    def ScrystalFun(T):

        sum_transformation_entropy = \
            totalTransformationEntropy(H_transformations, 0, T)
        integral = quad(fun, 0, T, limit=100)[0]
        crystal_entropy = integral + sum_transformation_entropy

        return crystal_entropy

    return ScrystalFun


def entropyGlassFun(Cp_glass_fun, glass_entropy_0K=0):
    '''
    Generate a function to compute the entropy of a glass.

    Parameters
    ----------
    Cp_glass_fun : callable
        Function with one argument for the temperature (in Kelvin) that returns
        the heat capacity at constant pressure for the glassy phase. It is
        necessary that Cp_glass_fun returns finite heat capacity values between
        zero Kelvin and any temperature that will be considerade to compute the
        entropy of the glass.

    glass_entropy_0K : integer or float
        The residual entropy of a glass at zero Kelvin. There are some debate
        in the literature about which is the residual entropy of a glass at
        this temperature. The default value is zero.

    Returns
    -------
    SglassFun : callable
        Function with one argument for temperature (in Kelvin) that returns the
        entropy of the glassy phase at the given temperature. The argument of
        the function can be a scalar or an array-like.

    '''
    def fun(x):
        return Cp_glass_fun(x) / x

    @np.vectorize
    def SglassFun(T):
        integral = quad(fun, 0, T, limit=100)[0]
        glass_entropy = integral + glass_entropy_0K

        return glass_entropy

    return SglassFun


def entropyLiquidFun(Cp_liquid_fun, Tm, delta_Hm, entropy_crystal_at_Tm):
    '''
    Generate a function to compute the entropy of a liquid.

    Parameters
    ----------
    Cp_liquid_fun : callable
        Function with one argument for the temperature (in Kelvin) that returns
        the heat capacity at constant pressure for the liquid phase.

    Tm : integer or float
        Melting temperature.

    delta_Hm : integer or float
        Enthalpy of melting.

    entropy_crystal_at_Tm : integer or float
        Entropy of the crystalline phas at the melting temperature right before
        the phase transition to the liquid.

    Returns
    -------
    SliquidFun : callable
        Function with one argument for temperature (in Kelvin) that returns the
        entropy of the liquid phase at the given temperature. The argument of
        the function can be a scalar or an array-like.

    '''
    delta_Sm = delta_Hm / Tm

    def fun(x):
        return Cp_liquid_fun(x) / x

    @np.vectorize
    def SliquidFun(T):
        integral = quad(fun, Tm, T, limit=100)[0]
        liquid_entropy = entropy_crystal_at_Tm + delta_Sm + integral

        return liquid_entropy

    return SliquidFun


def excessEntropyFun(Cp_crystal_fun,
                     Cp_liquid_fun,
                     Tm,
                     delta_Hm,
                     H_transformations=None):
    '''
    Generate a function to compute the excess entropy.

    The excess entropy is the entropy of the liquid phase minus the entropy of
    the crystalline phase.

    Parameters
    ----------
    Cp_crystal_fun : callable
        Function with one argument for the temperature (in Kelvin) that returns
        the heat capacity at constant pressure for the crystalline phase.

    Cp_liquid_fun : callable
        Function with one argument for the temperature (in Kelvin) that returns
        the heat capacity at constant pressure for the liquid phase.

    Tm : integer or float
        Melting temperature.

    delta_Hm : integer or float
        Enthalpy of melting.

    H_transformations : dictionary or None or False
        If there are any crystalline phase transitions in the material, this
        information must be given as a dictionary were the keys are the
        temperature where the transition happens (in Kelvin) and the values are
        the enthalpy of transformations (in J/mol) during heating. See the
        documentation for the function totalTransformationEntropy for more
        details. Default value is None.

    Returns
    -------
    SexcFun : callable
        Function with one argument for temperature (in Kelvin) that returns the
        excess entropy at the given temperature. The excess entropy is the
        difference between the liquid and the crystalline entropy at the given
        temperature The argument of the function can be a scalar or an
        array-like.

    '''
    delta_Sm = delta_Hm / Tm

    def fun(x):
        return (Cp_liquid_fun(x) - Cp_crystal_fun(x)) / x

    @np.vectorize
    def SexcFun(T):

        sum_transformation_entropy = \
            totalTransformationEntropy(H_transformations, T, Tm)
        integral = quad(fun, Tm, T, limit=100)[0]
        excess_entropy = delta_Sm + integral + sum_transformation_entropy

        return excess_entropy

    return SexcFun


def KauzmannTemperature(Cp_crystal_fun,
                        Cp_liquid_fun,
                        Tm,
                        delta_Hm,
                        H_transformations=None,
                        return_Sexc_fun=False):
    '''
    Computes the Kauzmann temperature where the excess entropy is zero.

    Parameters
    ----------
    Cp_crystal_fun : callable
        Function with one argument for the temperature (in Kelvin) that returns
        the heat capacity at constant pressure for the crystalline phase.

    Cp_liquid_fun : callable
        Function with one argument for the temperature (in Kelvin) that returns
        the heat capacity at constant pressure for the liquid phase.

    Tm : integer or float
        Melting temperature.

    delta_Hm : integer or float
        Enthalpy of melting.

    H_transformations : dictionary or None or False
        If there are any crystalline phase transitions in the material, this
        information must be given as a dictionary were the keys are the
        temperature where the transition happens (in Kelvin) and the values are
        the enthalpy of transformations (in J/mol) during heating. See the
        documentation for the function totalTransformationEntropy for more
        details. Default value is None.

    return_Sexc_fun : boolean
        If True then the excess entropy function is also returned. Default
        value is False.

    Returns
    -------
    TK : float or nan
        If there is a Kauzmann temperature then this function returns its value
        as a float. If no Kauzmann temperature is found, then the function
        returns nan.

    SexcFun : callable
        Only returned if 'return_Sexc_fun' is True. Function with one argument
        for temperature (in Kelvin) that returns the excess entropy at the
        given temperature. The excess entropy is the difference between the
        liquid and the crystalline entropy at the given temperature The
        argument of the function can be a scalar or an array-like.

    Notes
    -----
    There may be more than one temperature where the excess entropy is zero,
    depending on the crystalline phase transitions that are present. This
    function returns the highest Kauzmann temperature if there is more than one.

    References
    ----------
    [1] Kauzmann, W. (1948). The nature of the glassy state and the behavior of
        liquids at low temperatures. Chemical Reviews 43, 219–256.

    [2] Debenedetti, P.G., and Stillinger, F.H. (2001). Supercooled liquids and
        the glass transition. Nature 410, 259–267.

    '''
    SexcFun = excessEntropyFun(Cp_crystal_fun, Cp_liquid_fun, Tm, delta_Hm,
                               H_transformations)

    temperatures = np.linspace(1, Tm, max(15, int(Tm / 100)))

    if Tm > 300:
        temperatures = np.append(temperatures, 300)
        temperatures = np.sort(temperatures)

    if H_transformations:
        Tlist = []
        for transformation_T in H_transformations:
            Tlist.append(transformation_T - 1)
            Tlist.append(transformation_T + 1)

        temperatures = np.append(temperatures, Tlist)

    temperatures = np.sort(temperatures)

    Sexc = SexcFun(temperatures)
    logic1 = np.isfinite(Sexc)
    logic2 = Sexc != 0
    logic = np.logical_and(logic1, logic2)
    temperatures, Sexc = temperatures[logic], Sexc[logic]

    sign_change = np.where(np.diff(np.sign(Sexc)) != 0)[0]

    if len(sign_change) > 0:
        index = sign_change[-1]
        TK = brentq(SexcFun, temperatures[index], temperatures[index + 1])
    else:
        TK = np.nan

    if return_Sexc_fun:
        return TK, SexcFun
    else:
        return TK



import numpy as np


def _getCFunction(info):

    if info is None:

        def Cfun(T):
            return np.nan

    elif isinstance(info, dict):

        T0 = info.get('T0', 0)
        T1 = info.get('T1', 0)
        T2 = info.get('T2', 0)
        T3 = info.get('T3', 0)
        T4 = info.get('T4', 0)
        _T1 = info.get('_T1', 0)
        _T2 = info.get('_T2', 0)
        _T3 = info.get('_T3', 0)
        _T1_2 = info.get('_T1_2', 0)

        def Cfun(T_):
            T = float(T_)
            return T0 + T1 * T + T2 * T**2 + T3 * T**3 + T4 * T**4 \
                + _T1 * T**(-1) + _T2 * T**(-2) + _T3 * T**(-3) \
                + _T1_2 * T**(-1/2)

    else:
        raise TypeError('A dictionary or None must be provided as an argument')

    return Cfun


def heatCapacityFunFromTuple(C_tuple, extrapolate_borders=False):
    '''
    Generate a heat capacity function from a heat capacity tuple.

    Parameters
    ----------
    C_tuple : tuple
        Heat capacity tuple. See notes for the expected format.

    extrapolate_borders : True, False, 'both', 'upper', or 'lower'
        Parameter configuring if the generated function will extrapolate or not
        the functions in the C_tuple. If 'False' then no extrapolation will be
        made. If 'True' or 'both', then both temperature ends will be
        extrapolated if necessary. If 'lower' then only the heat capacity for
        lower temperatures will be extrapolated. If 'upper' then only the heat
        capacity for higher temperatures will be extrapolated. The default
        value is 'False'.

    Returns
    -------
    Cfun : callable
        Python function to compute the heat capacity following the information
        provided in C_tuple. This function accepts one argument named 'T' for
        temperature. 'T' can be any finite number or an array-like object. If
        the heat capacity cannot be computed (see 'extrapolate_borders') a nan
        value is returned.

    Notes
    -----
    The expected C_tuple format is a tuple of tuples. Below is a simple example
    for the heat capacity of liquid B2O3:

        Cp_tuple_B2O3_liquid = (
            (0, None),
            (723, {
                'T0': 128.3,
                'T1': -3.01E-4,
            }, 'liquid'),
            (3200, None),
        )

    Observe that the tuple Cp_tuple_B2O3_liquid has three tuples inside, each
    of these tuples with two or three items. These tuples inside the main tuple
    will be called subtuples from now on.

    The first item of these subtuples must be a float, indicating the lowest
    temperature limit of valitity of the heat capacity function of the
    subtuple. It is expected that the first subtuple in any C_tuple starts with
    zero Kelvin, as this is the lowest temperature where we can measure heat
    capacity. The upper temperature limit is always defined by the first item
    in the next subtuple in the sequence.

    The second item of the subtuples can be None or a dictionary with
    information about the heat capacity function. If the second item is None,
    then there is no heat capacity function defined. If the second item is a
    dictionary, then the keys of the dictionary indicate which coefficients of
    the heat capacity function are present, and the values of the dictionary
    indicate the values of the coefficients. In the example, in the range of
    723 to 3200 Kelvin, there are two coefficients, namely 'T0' and 'T1'. In
    this particular case, the heat capacity function is equivalent to:

    def function(T):
        return -3.01e-4*T + 128.3

    The valid coefficient names are:

        'T0': coefficient independent of temperature
        'T1': coefficient that multiplies T
        'T2': coefficient that multiplies T**2
        'T3': coefficient that multiplies T**3
        'T4': coefficient that multiplies T**4
        '_T1': coefficient that multiplies T**(-1)
        '_T2': coefficient that multiplies T**(-2)
        '_T3': coefficient that multiplies T**(-3)
        '_T1_2': coefficient that multiplies T**(-1/2)

    These are the usual coefficients found in the literature. More coefficients
    can be easily added by modifying '_getCFunction'.

    Finally, the third and last item of the subtuple is optional. It can hold
    any information that the user wants to store. It is never checked in any
    function of GlassPy.

    References
    ----------
    [1] Maier, C.G., and Kelley, K. (1932). An equation for the representation
        of high-temperature heat content data. Journal of the American Chemical
        Society 54, 3243–3246.

    '''
    functions = []

    if extrapolate_borders is False:
        for i in range(len(C_tuple)):
            segment_function = _getCFunction(C_tuple[i][1])
            functions.append(segment_function)

    else:
        if C_tuple[0][1] is None and (
                extrapolate_borders is True
                or extrapolate_borders in ['lower', 'both']):
            segment_function = _getCFunction(C_tuple[1][1])
            functions.append(segment_function)
        else:
            segment_function = _getCFunction(C_tuple[0][1])
            functions.append(segment_function)

        for i in range(len(C_tuple) - 2):
            segment_function = _getCFunction(C_tuple[i + 1][1])
            functions.append(segment_function)

        if C_tuple[-1][1] is None and (
                extrapolate_borders is True
                or extrapolate_borders in ['upper', 'both']):
            segment_function = _getCFunction(C_tuple[-2][1])
            functions.append(segment_function)
        else:
            segment_function = _getCFunction(C_tuple[-1][1])
            functions.append(segment_function)

    @np.vectorize
    def Cfun(T):
        if T < 0:
            return np.nan

        elif T == 0:
            return 0

        else:
            for i in range(len(C_tuple) - 1):
                if T > C_tuple[i][0] and T <= C_tuple[i + 1][0]:
                    return functions[i](T)

            return functions[-1](T)

    return Cfun



import numpy as np
from numpy import exp


def wakeshima(time, steady_state_rate, time_lag):
    """
    Computes the time-dependent nucleation rate using the Wakeshima equation.

    Parameters
    ----------
    time : float or array_like
        Elapsed time.

    steady_state_rate : float
        Steady-state nucleation rate.

    time_lag : float
        Time-lag of transient nucleation. See ref. [1].

    Returns
    -------
    nucleation_rate : float or array_like
        Returns the nucleation rate.

    References
    ----------
    [1] Wakeshima, H. (1954). Time Lag in the Self‐Nucleation. The Journal of
        Chemical Physics 22, 1614–1615.

    [2] Wakeshima, H. (1954). Fog Formation due to Self Nucleation. J. Phys.
        Soc. Jpn. 9, 400–406.

    [3] Wakeshima, H. (1955). Development of Emulsions due to Self Nucleation.
        Journal of the Physical Society of Japan 10, 65–70.

    [4] Wakeshima, H. (1955). Time Lag in the Self-Nucleation. J. Phys. Soc.
        Jpn. 10, 374–380.

    [5] Wakeshima, H. (1955). Errata: Time Lag in the Self-Nucleation. The
        Journal of Chemical Physics 23, 763–763.

    """
    time_ratio = time / time_lag
    nucleation_rate = steady_state_rate * (1 - exp(-time_ratio))
    return nucleation_rate


def kashchiev(time,
              steady_state_rate,
              time_lag,
              time_shift=0,
              summation_ub=1000):
    """
    Computes the time-dependent nucleation rate using the Kashchiev equation.

    Parameters
    ----------
    time : float or array_like
        Elapsed time.

    steady_state_rate : float
        Steady-state nucleation rate.

    time_lag : float
        Time-lag of transient nucleation. See ref. [1].

    time_shift : float, optional
        Time-shift to account for double-stage treatments. See Eq. (33.66) in
        ref. [2]. Default value is 0.

    summation_ub : int, optional
        Upper boundary of the infinite summation. Default value is 1000. It is
        advisable to choose an even integer.

    Returns
    -------
    out : float or array_like
        Returns the nucleation rate.

    Notes
    -----
    This is the expression with the time-shift factor. For the original
    expression deduced in Ref. [1], set time_shift to zero.

    References
    ----------
    [1] Kashchiev, D. (1969). Solution of the non-steady state problem in
        nucleation kinetics. Surface Science 14, 209–220.

    [2] Kashchiev, D. (2000). Nucleation basic theory with applications.
    """
    @np.vectorize
    def _kashchiev(t):
        if t <= time_shift:
            return 0

        try:
            time_ratio = (t - time_shift) / time_lag
        except ZeroDivisionError:
            time_ratio = np.inf

        def summationParticle(n):
            return ((-1)**(n % 2)) * exp(-n**2 * time_ratio)

        summation = np.sum(summationParticle(np.arange(1, summation_ub)))
        I = steady_state_rate * (1 + 2 * summation)
        return I if I > 0 else 0

    nucleation_rate = _kashchiev(time)
    return nucleation_rate


def kashchievMasterCurve(time_ratio, summation_ub=1000):
    """
    Computes the normalized nucleation rate using the Kashchiev equation.

    Parameters
    ----------
    time_ratio : float or array_like
        The ratio between the difference of time and time-shift over time-lag.

    summation_ub : int, optional
        Upper boundary of the infinite summation. Default value is 1000. It is
        advisable to choose an even integer.

    Returns
    -------
    normalized_nucleation_rate : float or array_like
        Returns the ration between the nucleation rate and the steady-state
        nucleation rate.

    References
    ----------
    [1] Kashchiev, D. (1969). Solution of the non-steady state problem in
        nucleation kinetics. Surface Science 14, 209–220.

    [2] Kashchiev, D. (2000). Nucleation basic theory with applications.
    """
    @np.vectorize
    def _kashchiev(time_ratio):

        if time_ratio < 0.1:
            return 0.
        elif time_ratio > 20:
            return 1.

        def summationParticle(n):
            return ((-1)**(n % 2)) * exp(-n**2 * time_ratio)

        summation = np.sum(summationParticle(np.arange(1, summation_ub)))
        normalized_I = 1 + 2 * summation
        return normalized_I if normalized_I > 0 else 0

    normalized_nucleation_rate = _kashchiev(time_ratio)
    return normalized_nucleation_rate


def shneidman(time, steady_state_rate, time_lag, time_incubation):
    """
    Computes the time-dependent nucleation rate using the Shneidman equation.

    Parameters
    ----------
    time : float or array_like
        Elapsed time.

    steady_state_rate : float
        Steady-state nucleation rate.

    time_lag : float
        Time-lag of transient nucleation. See Ref. [1].

    time_incubation : float
        Incubation time due to the double-stage treatment. See Ref. [1].

    Returns
    -------
    out : float or array_like
        Returns the nucleation rate.

    References
    ----------
    [1] Shneidman, V.A. (1988). Establishment of a steady-state nucleation
        regime. Theory and comparison with experimental data for glasses. Sov.
        Phys. Tech. Phys. 33, 1338–1342.
    """
    time_ratio = (time - time_incubation) / time_lag
    nucleation_rate = steady_state_rate * exp(-exp(-time_ratio))
    return nucleation_rate



'''Functions related to the Kohlrausch equation for relaxation, also known as
stretched exponential.'''

from math import gamma
from numpy import exp


def kohlrausch(time, tau_k, beta):
    '''Stretched exponential function commonly used for relaxation studies

    Parameters
    ----------
    time : float or array_like
        Elapsed time. The unity must be the same as tau_k.

    tau_k : float
        Characteristic time for the relaxation process. The unity must be the
        same as time.

    beta : float
        Stretched exponent. Must be greater than zero. Usually less than or
        equal to 1.

    Returns
    -------
    relaxation_parameter : float or array_like
        Number between zero and one indicating the degree of relaxation of the
        system. A value of one is for a system which is not relaxed. A value of
        zero is for a system fully relaxed.

    References
    ----------
    [1] Kohlrausch, R.H.A. (1854). Theorie des elektrischen Rückstandes in der
        Leidener Flasche. Annalen Der Physik Und Chemie 167, 179–214.

    [2] Ngai, K.L. (2011). Relaxation and Diffusion in Complex Systems (New
        York: Springer).

    '''
    relaxation_parameter = exp(-(time/tau_k)**beta)
    return relaxation_parameter


def tau_k2tau_ave(tau_k, beta):
    '''Converts the characteristic Kohlrausch time to the average relax. time

    Parameters
    ----------
    tau_k : float
        Characteristic time for the relaxation process.

    beta : float
        Stretched exponent. Must be greater than zero. Usually less than or
        equal to 1.

    Returns
    -------
    tau_ave : float
        Average relaxation time.

    '''
    tau_ave = tau_k*gamma(1/beta + 1)
    return tau_ave


def tau_ave2tau_k(tau_ave, beta):
    '''Converts the average relax. time to the characteristic Kohlrausch time

    Parameters
    ----------
    tau_ave : float
        Average relaxation time.

    beta : float
        Stretched exponent. Must be greater than zero. Usually less than or
        equal to 1.

    Returns
    -------
    tau_k : float
        Characteristic time for the relaxation process. 
    '''
    tau_k = tauAve/(gamma(1/beta + 1))
    return tau_k


def propertyRelaxation(time, tau_k, beta, p_0, p_inf):
    '''Computes the time-dependent property relaxation kinetics.

    Parameters
    ----------
    time : float or array_like
        Elapsed time. The unity must be the same as tau_k.

    tau_k : float
        Characteristic time for the relaxation process. The unity must be the
        same as time.

    beta : float
        Stretched exponent. Must be greater than zero. Usually less than or
        equal to 1.

    p_0 : float 
        Value of a certain property of the system at time equal zero.
        
    p_inf : float 
        Value of a certain property of the system at the limit of time
        approaching infinity. In other words, the value of the property for a
        fully relaxed system.

    Returns
    -------
    property_time : float or array_like
        Value of a certain property at different times during the relaxation
        process.

    References
    ----------
    [1] Kohlrausch, R.H.A. (1854). Theorie des elektrischen Rückstandes in der
        Leidener Flasche. Annalen Der Physik Und Chemie 167, 179–214.

    [2] Ngai, K.L. (2011). Relaxation and Diffusion in Complex Systems (New
        York: Springer).

    '''
    property_time = kohlrausch(time, tau_k, beta)*(p_0 - p_inf) + p_inf
    return property_time





import setuptools

with open("README_PyPI.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='glasspy',
    version='0.1',
    author='Daniel Roberto Cassar',
    author_email='daniel.r.cassar@gmail.com',
    description='Python module for scientists working with glass materials',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/drcassar/glasspy",
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy>=1.1', 'scipy>=0.19', 'pandas>=0.24.0', 'lmfit>=0.9.13'
    ],
    keywords='glass, non-crystalline materials',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha",
        "Topic :: Scientific/Engineering",
        "Intended Audience :: Science/Research",
        "Environment :: Console",
    ],
    license='GPL',
    python_requires='>=3.6',
)
